#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          14-08-2018
//      Created for Project :   FLAME
"""


# Python standard libs
from os import listdir
from os.path import isfile, join, dirname, splitext

# PIP installed libs
import pytest
from pyramid.httpexceptions import HTTPBadRequest
from yaml import load
from pyramid import testing
from requests import get, delete
from toscaparser.tosca_template import ToscaTemplate

# CLMC-service imports
from clmcservice.alertsapi.utilities import adjust_tosca_definitions_import, SFEMC
from clmcservice.alertsapi.alerts_specification_schema import validate_clmc_alerts_specification
from clmcservice.alertsapi.views import AlertsConfigurationAPI
from clmcservice import ROOT_DIR


class TestAlertsConfigurationAPI(object):
    """
    A pytest-implementation test for the Alerts Configuration API endpoints.
    """

    @pytest.fixture(autouse=True)
    def print_fixture(self):
        """
        Fixture to adjust the printing format when running pytest with the "-s" flag - by default print messages mix up with pytest's output
        """

        print()

    @pytest.fixture()
    def app_config(self):
        """
        A fixture to implement setUp/tearDown functionality for all tests by initializing configuration structure for the web service
        """

        self.config = testing.setUp()
        self.config.add_settings({"kapacitor_host": "localhost", "kapacitor_port": 9092, "sfemc_fqdn": "sfemc.localhost", "sfemc_port": 8081})

        yield

        testing.tearDown()

    @pytest.mark.parametrize("sfc_id, sfci_id, policy_id, trigger_id, task_id, topic_id", [
        ("MSDemo", "MSDemo-premium", "requests_diff", "low_requests", "094f23d6e948c78e9fa215528973fb3aeefa5525898626c9ea049dc8e87a7388", "094f23d6e948c78e9fa215528973fb3aeefa5525898626c9ea049dc8e87a7388"),
        ("sfc", "sfc_1", "missing_measurements", "no_latency_measurements", "1019ab70402a98905ff9e950c36d3c13a8468db8f7b8196f4ee69d6c39947fe5", "1019ab70402a98905ff9e950c36d3c13a8468db8f7b8196f4ee69d6c39947fe5"),
        ("disney.sfc.flame.eu", "disney.sfc.flame.eu-premium", "rtt_deviation", "increase_in_rtt", "8b069cc10571f887ff6c3a14aea11e68e664ddde5bc6af08d88058865510af93", "8b069cc10571f887ff6c3a14aea11e68e664ddde5bc6af08d88058865510af93"),
        ("vrt/novel/sfc", "vrt/novel/sfc-prototype", "deadmen_measurement", "missing_storage_measurements", "8c1aba1c61c184809d648e4ab323b9b48606847bb2e3d54bfbdf34d8a7db91d1", "8c1aba1c61c184809d648e4ab323b9b48606847bb2e3d54bfbdf34d8a7db91d1"),
        ("vrt/novel/sfc", "vrt/novel/sfc-deployment", "deadmen_measurement", "missing_storage_measurements", "a61dbe20cbef19be384982b51e22d83f23c4e10aeede172ef8646c515b836f7b", "a61dbe20cbef19be384982b51e22d83f23c4e10aeede172ef8646c515b836f7b")
    ])
    def test_alerts_hash_getter(self, sfc_id, sfci_id, policy_id, trigger_id, task_id, topic_id):
        """
        Tests the GET method for retrieving the task and topic ID given the sfc, sfc instance, policy id and trigger id.
        """

        request = testing.DummyRequest()
        request.params["sfc"] = sfc_id
        request.params["sfci"] = sfci_id
        request.params["policy"] = policy_id
        request.params["trigger"] = trigger_id

        response = AlertsConfigurationAPI(request).get_alerts_hash()

        assert response["task_identifier"] == task_id, "Incorrect task identifier returned"
        assert response["task_api_endpoint"] == "/kapacitor/v1/tasks/{0}".format(task_id), "Incorrect task API endpoint returned."
        assert response["topic_identifier"] == topic_id, "Incorrect topic identifier returned"
        assert response["topic_api_endpoint"] == "/kapacitor/v1/alerts/topics/{0}".format(topic_id), "Incorrect topic API endpoint returned"
        assert response["topic_handlers_api_endpoint"] == "/kapacitor/v1/alerts/topics/{0}/handlers".format(topic_id), "Incorrect topic handlers API endpoint returned"

    def test_alerts_config_tosca_parsing(self):
        """
        Tests that what we consider a valid/invalid alerts specification is successfully/unsuccessfully parsed by the TOSCA-parser.
        """

        for path_suffix, valid_expected in (("valid", True), ("invalid", False)):
            test_data_path = join(dirname(ROOT_DIR), *["resources", "tosca", "test-data", "tosca-parser", path_suffix])

            for test_file_path in listdir(test_data_path):
                alert_config_abs_path = join(test_data_path, test_file_path)

                if not isfile(alert_config_abs_path):
                    continue  # skip directories

                print(alert_config_abs_path, valid_expected)

                with open(alert_config_abs_path, 'r') as fh:
                    yaml_content = load(fh)
                    adjust_tosca_definitions_import(yaml_content)

                valid_real = True
                try:
                    ToscaTemplate(yaml_dict_tpl=yaml_content)
                except Exception:
                    valid_real = False

                assert valid_expected == valid_real, "TOSCA parser test failed for file: {0}".format(alert_config_abs_path)

    def test_alerts_config_clmc_validation(self):
        """
        Tests the custom CLMC validation of the TOSCA alerts specification.
        """

        for path_suffix, valid_expected in (("valid", True), ("invalid", False)):
            test_data_path = join(dirname(ROOT_DIR), *["resources", "tosca", "test-data", "clmc-validator", path_suffix])

            for test_file_path in listdir(test_data_path):
                alert_config_abs_path = join(test_data_path, test_file_path)

                if not isfile(alert_config_abs_path):
                    continue  # skip directories

                print(alert_config_abs_path, valid_expected)

                with open(alert_config_abs_path, 'r') as fh:
                    yaml_content = load(fh)
                    adjust_tosca_definitions_import(yaml_content)

                # do not catch exceptions here since we are testing the clmc validator, the tosca parsing is tested in the previous test method
                alert_tosca_spec = ToscaTemplate(yaml_dict_tpl=yaml_content)
                valid_real, err = validate_clmc_alerts_specification(alert_tosca_spec.tpl, include_error=True)
                assert valid_expected == valid_real, "CLMC alerts specification validator test failed for file: {0}".format(alert_config_abs_path)

    def test_alerts_config_api_post(self, app_config):
        """
        Tests the POST API endpoint of the alerts configuration API responsible for creating alerts.

        Test steps are:
            * Traverse all valid TOSCA Alerts Specifications and TOSCA Resource Specifications in the
                src/service/clmcservice/resources/tosca/test-data/clmc-validator/valid and src/service/clmcservice/resources/tosca/test-data/tosca-parser/valid folders
            * Send a valid TOSCA Alert Specification to the view responsible for configuring Kapacitor and creating alerts (POST request)
            * Check that Kapacitor alerts, topics and handlers are created with the correct identifier and arguments
            * Check that the API returns the duplication errors if the same alerts specification is sent through a POST request
            * Check that the API returns no duplication errors if the same alerts specification is sent through a PUT request
            * Clean up the registered alerts

        :param app_config: fixture for setUp/tearDown of the web service registry
        """

        kapacitor_host = self.config.registry.settings["kapacitor_host"]
        kapacitor_port = self.config.registry.settings["kapacitor_port"]

        # test all of the test files provided by the path_generator_testfiles function
        for alert_spec_file_paths, valid_resource_spec_file_paths, invalid_resource_spec_file_paths in path_generator_testfiles():

            alert_spec_abs_path, alerts_test_file = alert_spec_file_paths  # absolute path and name of the alerts spec. file
            valid_resource_spec_abs_path, valid_resources_test_file = valid_resource_spec_file_paths  # absolute path and name of a valid resource spec. file
            invalid_resource_spec_abs_path, invalid_resources_test_file = invalid_resource_spec_file_paths  # absolute path and name of an invalid resource spec. file

            with open(alert_spec_abs_path) as alert_spec:
                # first send an inconsistent resource spec, expecting bad request
                with open(invalid_resource_spec_abs_path) as invalid_resource_spec:
                    request = testing.DummyRequest()
                    request.POST['alert-spec'] = FieldStorageMock(alerts_test_file, alert_spec)
                    request.POST['resource-spec'] = FieldStorageMock(invalid_resources_test_file, invalid_resource_spec)
                    try:
                        AlertsConfigurationAPI(request).post_alerts_specification()
                        assert False, "No error was returned even though an inconsistent resource specification was sent"
                    except HTTPBadRequest:
                        pass  # we expect this to happen

                # reset the read pointer of the alert specification file since it was already read once
                alert_spec.seek(0)

                # extract the alert specification data in a structured way
                sfc, sfc_instance, alerts = extract_alert_configuration_data(alert_spec, self.config.registry.settings["sfemc_fqdn"], self.config.registry.settings["sfemc_port"])
                alert_spec.seek(0)  # reset the read pointer to the beginning again (the extraction in the previous step had to read the full file)

                # then send a consistent resource spec
                with open(valid_resource_spec_abs_path) as valid_resource_spec:
                    request = testing.DummyRequest()
                    request.POST['alert-spec'] = FieldStorageMock(alerts_test_file, alert_spec)  # a simple mock class is used to mimic the FieldStorage class
                    request.POST['resource-spec'] = FieldStorageMock(valid_resources_test_file, valid_resource_spec)
                    clmc_service_response = AlertsConfigurationAPI(request).post_alerts_specification()

            assert (sfc, sfc_instance) == (clmc_service_response["service_function_chain_id"], clmc_service_response["service_function_chain_instance_id"]), \
                "Incorrect extraction of metadata for file {0}". format(alerts_test_file)
            assert "triggers_specification_errors" not in clmc_service_response, "Unexpected error was returned for triggers specification"
            assert "triggers_action_errors" not in clmc_service_response, "Unexpected error was returned for handlers specification"

            # traverse through all alerts and check that they are created along with their respective handlers within Kapacitor
            check_kapacitor_alerts(alerts, kapacitor_host, kapacitor_port, alerts_test_file)

            # send the same spec again to check that error messages are returned (because of ID duplication)
            with open(alert_spec_abs_path) as alert_spec:
                with open(valid_resource_spec_abs_path) as valid_resource_spec:
                    request.POST['alert-spec'] = FieldStorageMock(alerts_test_file, alert_spec)  # a simple mock class is used to mimic the FieldStorage class
                    request.POST['resource-spec'] = FieldStorageMock(valid_resources_test_file, valid_resource_spec)
                    clmc_service_response = AlertsConfigurationAPI(request).post_alerts_specification()
            assert (sfc, sfc_instance) == (clmc_service_response["service_function_chain_id"], clmc_service_response["service_function_chain_instance_id"]), \
                "Incorrect extraction of metadata for file {0}". format(alerts_test_file)

            assert len(clmc_service_response["triggers_specification_errors"]) == len(alerts), "Expected errors were not returned for triggers specification"
            handlers_count = sum([len(alert["handlers"]) for alert in alerts])
            assert len(clmc_service_response["triggers_action_errors"]) == handlers_count, "Expected errors were not returned for handlers specification"

            # send the same request but as a PUT method instead of POST
            with open(alert_spec_abs_path) as alert_spec:
                with open(valid_resource_spec_abs_path) as valid_resource_spec:
                    request.params['alert-spec'] = FieldStorageMock(alerts_test_file, alert_spec)  # a simple mock class is used to mimic the FieldStorage class
                    request.params['resource-spec'] = FieldStorageMock(valid_resources_test_file, valid_resource_spec)
                    clmc_service_response = AlertsConfigurationAPI(request).put_alerts_specification()

            # no errors are expected now, since the PUT request must update existing alerts
            assert (sfc, sfc_instance) == (clmc_service_response["service_function_chain_id"], clmc_service_response["service_function_chain_instance_id"]), \
                "Incorrect extraction of metadata for file {0}". format(alerts_test_file)
            assert "triggers_specification_errors" not in clmc_service_response, "Unexpected error was returned for triggers specification"
            assert "triggers_action_errors" not in clmc_service_response, "Unexpected error was returned for handlers specification"

            # traverse through all alerts and check that they were not deleted by the PUT request
            check_kapacitor_alerts(alerts, kapacitor_host, kapacitor_port, alerts_test_file)

            # clean-up in the end of the test
            clean_kapacitor_alerts(alerts, kapacitor_host, kapacitor_port)

    def test_alerts_config_api_put(self, app_config):
        """
        Tests the PUT API endpoint of the alerts configuration API responsible for creating or updating alerts.

        Test steps are:
            * Traverse all valid TOSCA Alerts Specifications and TOSCA Resource Specifications in the
                src/service/clmcservice/resources/tosca/test-data/clmc-validator/valid and src/service/clmcservice/resources/tosca/test-data/tosca-parser/valid folders
            * Send a valid TOSCA Alert Specification to the view responsible for configuring Kapacitor and creating/updating alerts (PUT request)
            * Check that Kapacitor alerts, topics and handlers are created with the correct identifier and arguments
            * Clean up the registered alerts

        :param app_config: fixture for setUp/tearDown of the web service registry
        """

        kapacitor_host = self.config.registry.settings["kapacitor_host"]
        kapacitor_port = self.config.registry.settings["kapacitor_port"]

        # test all of the test files provided by the path_generator_testfiles function, ignoring the last result, which is invalid resource spec. files
        for alert_spec_file_paths, valid_resource_spec_file_paths, _ in path_generator_testfiles():
            alert_spec_abs_path, alerts_test_file = alert_spec_file_paths  # absolute path and name of the alerts spec. file
            resource_spec_abs_path, resources_test_file = valid_resource_spec_file_paths  # absolute path and name of a valid resource spec. file

            with open(alert_spec_abs_path) as alert_spec:
                # extract alert configuration data
                sfc, sfc_instance, alerts = extract_alert_configuration_data(alert_spec, self.config.registry.settings["sfemc_fqdn"], self.config.registry.settings["sfemc_port"])
                alert_spec.seek(0)  # reset the read pointer to the beginning again (the extraction in the previous step had to read the full file)

                # send valid alert and resource spec to create the alerts with a PUT request (can be used for updating or creating)
                with open(resource_spec_abs_path) as resource_spec:
                    request = testing.DummyRequest()
                    request.params['alert-spec'] = FieldStorageMock(alerts_test_file, alert_spec)  # a simple mock class is used to mimic the FieldStorage class
                    request.params['resource-spec'] = FieldStorageMock(resources_test_file, resource_spec)
                    clmc_service_response = AlertsConfigurationAPI(request).put_alerts_specification()

            # no errors are expected now, since the PUT request must update existing alerts
            assert (sfc, sfc_instance) == (clmc_service_response["service_function_chain_id"], clmc_service_response["service_function_chain_instance_id"]), \
                "Incorrect extraction of metadata for file {0}".format(alerts_test_file)
            assert "triggers_specification_errors" not in clmc_service_response, "Unexpected error was returned for triggers specification"
            assert "triggers_action_errors" not in clmc_service_response, "Unexpected error was returned for handlers specification"

            # traverse through all alerts and check that they were created by the PUT request
            check_kapacitor_alerts(alerts, kapacitor_host, kapacitor_port, alerts_test_file)

            # clean-up in the end of the test
            clean_kapacitor_alerts(alerts, kapacitor_host, kapacitor_port)

    def test_alerts_config_api_get(self, app_config):
        """
        Tests the GET API endpoint of the alerts configuration API responsible for fetching registered alerts for a specific SFC instance.

        Test steps are:
            * Traverse all valid TOSCA Alerts Specifications and TOSCA Resource Specifications in the
                src/service/clmcservice/resources/tosca/test-data/clmc-validator/valid and src/service/clmcservice/resources/tosca/test-data/tosca-parser/valid folders
            * Send a valid TOSCA Alert Specification to the view responsible for configuring Kapacitor and creating alerts
            * Send a valid TOSCA Alert Specification to the view responsible for getting the created alerts
            * Check that all alerts and Kapacitor resources (task, topic, handler urls) have been correctly returned
            * Clean up the alerts

        :param app_config: fixture for setUp/tearDown of the web service registry
        """

        kapacitor_host = self.config.registry.settings["kapacitor_host"]
        kapacitor_port = self.config.registry.settings["kapacitor_port"]
        sfemc_fqdn = self.config.registry.settings["sfemc_fqdn"]
        sfemc_port = self.config.registry.settings["sfemc_port"]
        fqdn_prefix = "http://{0}:{1}/sfemc/event".format(sfemc_fqdn, sfemc_port)

        # test all of the test files provided by the path_generator_testfiles function, ignoring the last result, which is invalid resource spec. files
        for alert_spec_file_paths, valid_resource_spec_file_paths, _ in path_generator_testfiles():

            alert_spec_abs_path, alerts_test_file = alert_spec_file_paths  # absolute path and name of the alerts spec. file
            resource_spec_abs_path, resources_test_file = valid_resource_spec_file_paths  # absolute path and name of a valid resource spec. file

            with open(alert_spec_abs_path) as alert_spec:

                # extract alert configuration data
                sfc, sfc_instance, alerts = extract_alert_configuration_data(alert_spec, self.config.registry.settings["sfemc_fqdn"], self.config.registry.settings["sfemc_port"])
                alert_spec.seek(0)  # reset the read pointer to the beginning again (the extraction in the previous step had to read the full file)

                # send a GET request for registered alerts, expecting an empty lists, alerts have not been created yet
                request = testing.DummyRequest()
                request.matchdict["sfc_id"] = sfc
                request.matchdict["sfc_instance_id"] = sfc_instance
                response = AlertsConfigurationAPI(request).get_alerts()
                assert response == [], "Incorrect response when fetching registered alerts, expecting empty list"

                # send valid alert and resource spec to create the alerts to be fetched afterwards through a GET request
                with open(resource_spec_abs_path) as resource_spec:
                    request = testing.DummyRequest()
                    request.POST['alert-spec'] = FieldStorageMock(alerts_test_file, alert_spec)  # a simple mock class is used to mimic the FieldStorage class
                    request.POST['resource-spec'] = FieldStorageMock(resources_test_file, resource_spec)
                    AlertsConfigurationAPI(request).post_alerts_specification()

            # send a GET request for registered alerts, expecting the newly created alerts
            request = testing.DummyRequest()
            request.matchdict["sfc_id"] = sfc
            request.matchdict["sfc_instance_id"] = sfc_instance
            response = AlertsConfigurationAPI(request).get_alerts()

            # restructure the extracted alerts data to be comparable with the response of the clmc service
            expected_alerts = map(lambda alert_object: {
                "policy": alert_object["policy"], "trigger": alert_object["trigger"],
                # make sure handlers are sorted and replace URLs starting with the SFEMC fqdn prefix to the SFEMC label
                "handlers": sorted(map(lambda handler: SFEMC if handler.startswith(fqdn_prefix) else handler, alert_object["handlers"].values())),
                "task_identifier": alert_object["task"], "topic_identifier": alert_object["topic"],
                "task_api_endpoint": "/kapacitor/v1/tasks/{0}".format(alert_object["task"]),
                "topic_api_endpoint": "/kapacitor/v1/alerts/topics/{0}".format(alert_object["topic"]),
                "topic_handlers_api_endpoint": "/kapacitor/v1/alerts/topics/{0}/handlers".format(alert_object["topic"])
            }, alerts)
            expected_alerts = sorted(expected_alerts, key=lambda x: x["trigger"])
            # sort the handlers of each alert in the response to ensure comparison order is correct
            for alert in response:
                alert["handlers"] = sorted(alert["handlers"])

            # compare the actual response with the expected response
            assert sorted(response, key=lambda x: x["trigger"]) == expected_alerts, "Incorrect result returned from a GET alerts request"

            clean_kapacitor_alerts(alerts, kapacitor_host, kapacitor_port)

    def test_alerts_config_api_delete(self, app_config):
        """
        Tests the DELETE API endpoint of the alerts configuration API responsible for deleting alerts specifications.

        Test steps are:
            * Traverse all valid TOSCA Alerts Specifications and TOSCA Resource Specifications in the
                src/service/clmcservice/resources/tosca/test-data/clmc-validator/valid and src/service/clmcservice/resources/tosca/test-data/tosca-parser/valid folders
            * Send a valid TOSCA Alert Specification to the view responsible for configuring Kapacitor and creating alerts
            * Send a valid TOSCA Alert Specification to the view responsible for deleting the created alerts if they exist
            * Check that all Kapacitor resources (task, topic, handler) have been deleted

        :param app_config: fixture for setUp/tearDown of the web service registry
        """

        kapacitor_host = self.config.registry.settings["kapacitor_host"]
        kapacitor_port = self.config.registry.settings["kapacitor_port"]
        sfemc_fqdn = self.config.registry.settings["sfemc_fqdn"]
        sfemc_port = self.config.registry.settings["sfemc_port"]

        # test all of the test files provided by the path_generator_testfiles function, ignoring the last result, which is invalid resource spec. files
        for alert_spec_file_paths, valid_resource_spec_file_paths, _ in path_generator_testfiles():

            alert_spec_abs_path, alerts_test_file = alert_spec_file_paths  # absolute path and name of the alerts spec. file
            resource_spec_abs_path, resources_test_file = valid_resource_spec_file_paths  # absolute path and name of a valid resource spec. file

            with open(alert_spec_abs_path) as alert_spec:

                # extract the alert specification data, ignore SFC and SFC instance IDs
                _, _, alerts = extract_alert_configuration_data(alert_spec, sfemc_fqdn, sfemc_port)
                alert_spec.seek(0)  # reset the read pointer to the beginning again (the extraction in the previous step had to read the full file)

                # send valid alert and resource spec to create the alerts to be deleted afterwards
                with open(resource_spec_abs_path) as resource_spec:
                    request = testing.DummyRequest()
                    request.POST['alert-spec'] = FieldStorageMock(alerts_test_file, alert_spec)  # a simple mock class is used to mimic the FieldStorage class
                    request.POST['resource-spec'] = FieldStorageMock(resources_test_file, resource_spec)
                    AlertsConfigurationAPI(request).post_alerts_specification()

            # ensure that these resource (tasks, topics, handlers) exist in Kapacitor
            # traverse through all alerts and check that everything is created
            check_kapacitor_alerts(alerts, kapacitor_host, kapacitor_port, alerts_test_file)

            with open(alert_spec_abs_path) as alert_spec:
                # now send the alert spec for deletion and check that everything is deleted in Kapacitor
                request = testing.DummyRequest()
                request.params['alert-spec'] = FieldStorageMock(alerts_test_file, alert_spec)
                response = AlertsConfigurationAPI(request).delete_alerts_specification()

            # restructure the extracted alerts data to be comparable with the response of the clmc service
            deleted_alerts = map(lambda alert_object: {"policy": alert_object["policy"], "trigger": alert_object["trigger"]}, alerts)
            deleted_alerts = sorted(deleted_alerts, key=lambda x: x["trigger"])
            deleted_handlers = []
            for alert in alerts:
                policy = alert["policy"]
                trigger = alert["trigger"]
                for handler_id in alert["handlers"]:
                    handler_url = alert["handlers"][handler_id]
                    if handler_url.startswith("http://{0}:{1}/sfemc/event".format(sfemc_fqdn, sfemc_port)):
                        handler_url = SFEMC
                    deleted_handlers.append({"policy": policy, "trigger": trigger, "handler": handler_url})
            deleted_handlers = sorted(deleted_handlers, key=lambda x: (x["trigger"], x["handler"]))

            # assert the response is what's expected containing the deleted alerts and handlers
            assert response.keys() == {"deleted_alerts", "deleted_handlers"}, "Incorrect response format"
            assert sorted(response["deleted_alerts"], key=lambda x: x["trigger"]) == deleted_alerts, "Incorrect result for deleted alerts"
            assert sorted(response["deleted_handlers"], key=lambda x: (x["trigger"], x["handler"])) == deleted_handlers, "Incorrect result for deleted handlers"

            # ensure that these resource (tasks, topics, handlers) do not exist in Kapacitor anymore
            # traverse through all alerts and check that everything is deleted from Kapacitor
            for alert in alerts:
                # check the alert task
                alert_id = alert["task"]
                kapacitor_response = get("http://{0}:{1}/kapacitor/v1/tasks/{2}".format(kapacitor_host, kapacitor_port, alert_id))
                assert kapacitor_response.status_code == 404, "Alert with ID {0} was not deleted - test file {1}.".format(alert_id, alerts_test_file)

                # check the alert topic
                topic_id = alert["topic"]
                kapacitor_response = get("http://{0}:{1}/kapacitor/v1/alerts/topics/{2}".format(kapacitor_host, kapacitor_port, topic_id))
                assert kapacitor_response.status_code == 404, "Topic with ID {0} was not deleted - test file {1}".format(topic_id, alerts_test_file)

                # check handlers
                for handler_id in alert["handlers"]:
                    kapacitor_response = get("http://{0}:{1}/kapacitor/v1/alerts/topics/{2}/handlers/{3}".format(kapacitor_host, kapacitor_port, topic_id, handler_id))
                    assert kapacitor_response.status_code == 404, "Handler with ID {0} for topic with ID {1} was not deleted - test file {2}".format(handler_id, topic_id, alerts_test_file)

            # now send a second delete request to ensure that nothing else is deleted
            with open(alert_spec_abs_path) as alert_spec:
                request = testing.DummyRequest()
                request.params['alert-spec'] = FieldStorageMock(alerts_test_file, alert_spec)
                response = AlertsConfigurationAPI(request).delete_alerts_specification()

            assert response == {"deleted_alerts": [], "deleted_handlers": []}, "Incorrect response after a second delete"


# The implementation below is for utility methods and classes used in the tests' implementation.
class FieldStorageMock(object):

    def __init__(self, filename, file):
        """
        Used to mock the behaviour of the cgi.FieldStorage class - two attributes needed only to forward a file to the view.

        :param filename: file name
        :param file: file object
        """

        self.filename = filename
        self.file = file


def path_generator_testfiles():
    """
    A utility function which returns a generator object for traversing the valid CLMC alert specs and their respective valid/invalid resource specs

    :return: a generator object that yields a triple of file paths (valid_alert_spec_path, valid_resource_spec_path, invalid_resource_spec_path)
    """

    test_folder = "clmc-validator"
    alerts_test_data_path = join(dirname(ROOT_DIR), *["resources", "tosca", "test-data", test_folder, "valid"])
    resources_test_data_path = join(dirname(ROOT_DIR), *["resources", "tosca", "test-data", "resource-spec"])

    # traverse through all files in the clmc-validator/valid folder (expected to be valid TOSCA alert specifications)
    for alerts_test_file in listdir(alerts_test_data_path):
        alert_spec_abs_path = join(alerts_test_data_path, alerts_test_file)

        if not isfile(alert_spec_abs_path):
            continue  # skip directories

        print("Testing file {0} in folder {1}".format(alerts_test_file, test_folder))

        # the respective resource specification consistent with this alert spec. will have the same name, with "alerts"
        # being replaced by "resources_valid" for valid spec or "resources_invalid" for invalid spec
        valid_resources_test_file = alerts_test_file.replace("alerts", "resources_valid")
        invalid_resources_test_file = alerts_test_file.replace("alerts", "resources_invalid")
        valid_resource_spec_abs_path = join(resources_test_data_path, valid_resources_test_file)
        invalid_resource_spec_abs_path = join(resources_test_data_path, invalid_resources_test_file)

        print("Test uses resource spec. files {0} and {1}".format(valid_resources_test_file, invalid_resources_test_file))

        yield (alert_spec_abs_path, alerts_test_file), (valid_resource_spec_abs_path, valid_resources_test_file), (invalid_resource_spec_abs_path, invalid_resources_test_file)


def extract_alert_configuration_data(alert_spec, sfemc_fqdn, sfemc_port):
    """
    A utility function to extract the expected alert, handler and topic identifiers (Kapacitor resources) from a given alert specification.

    :param alert_spec: the alert specification file (file object)
    :param sfemc_fqdn: FQDN of SFEMC
    :param sfemc_port: port number of SFEMC

    :return: a list of alert objects containing policy ID, trigger ID, handlers, generated task/topic ID, alert type
    """

    version = splitext(alert_spec.name)[0].split("-")[-1]  # take the ending number of the alert spec file
    yaml_alert_spec = load(alert_spec)
    adjust_tosca_definitions_import(yaml_alert_spec)
    tosca_tpl = ToscaTemplate(yaml_dict_tpl=yaml_alert_spec)
    # TODO next release - uncomment
    # sfc, sfc_instance = tosca_tpl.tpl["metadata"]["sfc"], tosca_tpl.tpl["metadata"]["sfci"]
    sfc, sfc_instance = tosca_tpl.tpl["metadata"]["servicefunctionchain"], "{0}_1".format(tosca_tpl.tpl["metadata"]["servicefunctionchain"])

    alerts = []  # saves every alert object

    for policy in tosca_tpl.policies:
        policy_id = policy.name
        for trigger in policy.triggers:
            trigger_id = trigger.name
            event_type = trigger.trigger_tpl["event_type"]
            alert_period_integer = trigger.trigger_tpl["condition"]["granularity"]

            topic_id = "{0}\n{1}\n{2}\n{3}".format(sfc, sfc_instance, policy_id, trigger_id)
            topic_id = AlertsConfigurationAPI.get_hash(topic_id)

            alert_id = topic_id
            alert_type = get_alert_type(event_type, alert_period_integer)

            alert_handlers = {}
            for handler_url in trigger.trigger_tpl["action"]["implementation"]:

                if handler_url == SFEMC:
                    handler_url = "http://{0}:{1}/sfemc/event/{2}/{3}/{4}".format(sfemc_fqdn, sfemc_port, sfc, policy_id, "trigger_id_{0}".format(version))

                handler_id = "{0}\n{1}\n{2}\n{3}\n{4}".format(sfc, sfc_instance, policy_id, trigger_id, handler_url)
                handler_id = AlertsConfigurationAPI.get_hash(handler_id)
                alert_handlers[handler_id] = handler_url

            alerts.append({"policy": policy_id, "trigger": trigger_id, "task": alert_id, "topic": topic_id, "type": alert_type, "handlers": alert_handlers})

    return sfc, sfc_instance, alerts


def get_alert_type(event_type, alert_period):
    """
    Retrieve the alert type (stream ot batch) based on the event type and alert period.

    :param event_type: event type, e.g. threshold, relative, deadman, etc.
    :param alert_period: the alert period

    :return: "batch" or "stream"
    """

    if event_type in AlertsConfigurationAPI.DUAL_VERSION_TEMPLATES:
        if alert_period < AlertsConfigurationAPI.STREAM_PERIOD_LIMIT:
            return "stream"
        else:
            return "batch"
    else:
        events = {"relative": "batch", "deadman": "stream"}
        return events[event_type]


def check_kapacitor_alerts(alerts, kapacitor_host, kapacitor_port, alerts_test_file):
    """
    Check that all Kapactior resources are created with the expected values.

    :param alerts: the list of alert objects to test
    :param kapacitor_host: hostname of kapacitor
    :param kapacitor_port: port number of kapacitor
    :param alerts_test_file: name of the current test file
    """

    for alert in alerts:

        alert_id = alert["task"]
        alert_type = alert["type"]
        topic_id = alert["topic"]

        # check the Kapacitor task
        kapacitor_response = get("http://{0}:{1}/kapacitor/v1/tasks/{2}".format(kapacitor_host, kapacitor_port, alert_id))
        assert kapacitor_response.status_code == 200, "Alert with ID {0} was not created - test file {1}.".format(alert_id, alerts_test_file)
        kapacitor_response_json = kapacitor_response.json()
        assert "link" in kapacitor_response_json, "Incorrect response from kapacitor for alert with ID {0} - test file {1}".format(alert_id, alerts_test_file)
        assert kapacitor_response_json["status"] == "enabled", "Alert with ID {0} was created but is disabled - test file {1}".format(alert_id, alerts_test_file)
        assert kapacitor_response_json["executing"], "Alert with ID {0} was created and is enabled, but is not executing - test file {1}".format(alert_id, alerts_test_file)
        assert kapacitor_response_json["type"] == alert_type, "Alert with ID {0} was created with the wrong type - test file {1}".format(alert_id, alerts_test_file)

        # check the Kapacitor topic
        kapacitor_response = get("http://{0}:{1}/kapacitor/v1/alerts/topics/{2}".format(kapacitor_host, kapacitor_port, topic_id))
        assert kapacitor_response.status_code == 200, "Topic with ID {0} was not created - test file {1}".format(topic_id, alerts_test_file)
        kapacitor_response_json = kapacitor_response.json()
        assert kapacitor_response_json["id"] == topic_id, "Topic {0} was created with incorrect ID - test file {1}".format(topic_id, alerts_test_file)

        # check that all handler IDs were created and each of them is subscribed to the correct topic ID
        for handler_id in alert["handlers"]:
            handler_url = alert["handlers"][handler_id]
            kapacitor_response = get("http://{0}:{1}/kapacitor/v1/alerts/topics/{2}/handlers/{3}".format(kapacitor_host, kapacitor_port, topic_id, handler_id))
            assert kapacitor_response.status_code == 200, "Handler with ID {0} for topic with ID {1} doesn't exist - test file {2}".format(handler_id, topic_id, alerts_test_file)
            kapacitor_response_json = kapacitor_response.json()
            assert kapacitor_response_json["id"] == handler_id, "Incorrect ID of handler {0} in the Kapacitor response - test file {1}".format(handler_id, alerts_test_file)
            assert kapacitor_response_json["kind"] == "post", "Incorrect kind of handler {0} in the Kapacitor response - test file {1}".format(handler_id, alerts_test_file)
            assert kapacitor_response_json["options"]["url"] == handler_url, "Incorrect url of handler {0} in the Kapacitor response - test file {1}".format(handler_id, alerts_test_file)


def clean_kapacitor_alerts(alerts, kapacitor_host, kapacitor_port):
    """
    A utility function to clean up Kapacitor from the configured alerts, topics and handlers.

    :param alerts: the list of alert objects along with handlers and generated identifiers
    :param kapacitor_host: Kapacitor hostname
    :param kapacitor_port: Kapacitor port number
    """

    for alert in alerts:

        # delete the alert task
        alert_id = alert["task"]
        kapacitor_response = delete("http://{0}:{1}/kapacitor/v1/tasks/{2}".format(kapacitor_host, kapacitor_port, alert_id))  # delete alert
        assert kapacitor_response.status_code == 204

        # delete the handlers
        topic_id = alert["topic"]
        for handler_id in alert["handlers"]:
            kapacitor_response = delete("http://{0}:{1}/kapacitor/v1/alerts/topics/{2}/handlers/{3}".format(kapacitor_host, kapacitor_port, topic_id, handler_id))  # delete handler
            assert kapacitor_response.status_code == 204

        # delete the alert topic
        kapacitor_response = delete("http://{0}:{1}/kapacitor/v1/alerts/topics/{2}".format(kapacitor_host, kapacitor_port, topic_id))  # delete topic
        assert kapacitor_response.status_code == 204
