#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          04-07-2018
//      Created for Project :   FLAME
"""


from clmcservice.graphapi.utilities import validate_build_request_body, validate_monitor_request_body, RTT_CYPHER_QUERY_TEMPLATE, \
    build_network_graph, delete_network_graph, build_temporal_subgraph, delete_temporal_subgraph, validate_graph_rtt_params, find_node_with_possible_types, depth_first_search
from clmcservice.models import MonitoringProcess
from influxdb import InfluxDBClient
from py2neo import Graph
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound, HTTPServiceUnavailable, HTTPNotImplemented, HTTPInternalServerError
from pyramid.view import view_defaults, view_config
from requests import exceptions, get
from uuid import uuid4
from json import load, dumps
from subprocess import Popen
from psutil import Process, NoSuchProcess, TimeoutExpired
from logging import getLogger


log = getLogger('service_logger')


@view_defaults(renderer='json')
class GraphAPI(object):
    """
    A class-based view for  building temporal graphs and running graph queries.
    """

    def __init__(self, request):
        """
        Initialises the instance of the view with the request argument.

        :param request: client's call request
        """

        self.request = request

    @view_config(route_name='graph_build', request_method='POST')
    def build_temporal_graph(self):
        """
        An API endpoint to build a temporal graph and store it in neo4j based on the posted JSON query document.

        :raises HTTPBadRequest: if request body is not a valid JSON with the queries per service function

        :return: A JSON document containing metadata about the built graph (time range and uuid, which can then be reused for other API calls)
        """

        try:
            body = self.request.body.decode(self.request.charset)
            json_queries = validate_build_request_body(body)  # validate the content and receive a json dictionary object
        except AssertionError as e:
            raise HTTPBadRequest("Bad request content: {0}".format(e.args))

        graph = self.get_graph_reference()
        influx_client = InfluxDBClient(host=self.request.registry.settings['influx_host'], port=self.request.registry.settings['influx_port'], timeout=10)

        database_name = json_queries["service_function_chain"]
        if database_name not in [db["name"] for db in influx_client.get_list_database()]:
            raise HTTPBadRequest("Database for service function chain {0} not found.".format(database_name))

        from_timestamp = json_queries['from'] * 10**9
        to_timestamp = json_queries['to'] * 10**9

        request_id = str(uuid4())

        endpoints_names = build_temporal_subgraph(request_id, from_timestamp, to_timestamp, json_queries, graph, influx_client)

        json_response = {"database": database_name, 'graph': {"uuid": request_id, "endpoints": endpoints_names, "time_range": {"from": from_timestamp, "to": to_timestamp}}}

        return json_response

    @view_config(route_name='temporal_graph_manage', request_method='DELETE')
    def delete_temporal_graph(self):
        """
        An API endpoint to delete a temporal graph associated with a uuid generated by the CLMC service.

        :return: A JSON document containing the number of deleted nodes.

        :raises HTTPNotFound: if the request is not associated with any subgraph
        """

        graph_id = self.request.matchdict['graph_id']  # get the UUID of the subgraph from the URL
        graph = self.get_graph_reference()

        if graph.nodes.match("Reference", uuid=graph_id).first() is None:
            raise HTTPNotFound("No subgraph found associated with the request ID {0}".format(graph_id))

        number_of_deleted_nodes = delete_temporal_subgraph(graph, graph_id)
        return {"deleted": number_of_deleted_nodes}

    @view_config(route_name='full_graph_manage', request_method='DELETE')
    def delete_full_graph(self):
        """
        An API endpoint to delete a media service graph with a given SFC identifier.

        :return: A JSON document containing the number of deleted nodes.

        :raises HTTPNotFound: if there is no SFC node with the given identifier
        """

        sfc_id = self.request.matchdict['sfc_id']  # get the SFC identifier from the URL
        graph = self.get_graph_reference()

        # check if this SFC node exists
        sfc_node = graph.nodes.match("ServiceFunctionChain", name=sfc_id).first()
        if sfc_node is None:
            raise HTTPNotFound("No service function chain node found with identifier {0}".format(sfc_id))

        # delete all nodes that are linked to the SFC node
        count = 0
        for node in depth_first_search(graph, sfc_node):
            graph.delete(node)
            count += 1

        # delete any reference nodes for temporal graphs that are associated to this SFC identifier
        for node in graph.nodes.match("Reference", sfc=sfc_id):
            graph.delete(node)
            count += 1

        return {"deleted": count}

    @view_config(route_name='graph_algorithms_rtt', request_method='GET')
    def run_rtt_query(self):
        """
        An API endpoint to run the round trip time cypher query over the graph associated with a given request ID.

        :return: A JSON response with a list of forward latencies, reverse latencies and SF endpoint response time.

        :raises HTTPBadRequest: if the request URL doesn't contain the required URL query parameters
        :raises HTTPNotFound: if the request is not associated with any subgraph or the compute node / endpoint node doesn't exist
        """

        graph_id = self.request.matchdict['graph_id']  # get the UUID of the subgraph from the URL

        try:
            params = validate_graph_rtt_params(self.request.params)
        except AssertionError as e:
            raise HTTPBadRequest("Request URL format is incorrect: {0}".format(e.args))

        startpoint_node_label = params["startpoint"]
        endpoint_node_label = params["endpoint"]

        graph = self.get_graph_reference()

        all_nodes = graph.nodes

        reference_node = all_nodes.match("Reference", uuid=graph_id).first()
        if reference_node is None:
            raise HTTPNotFound("No subgraph found associated with the request ID {0}".format(graph_id))

        # match a switch, cluster or ue node as a path start point and capture the type of the node
        startpoint_node, startpoint_node_type = find_node_with_possible_types(startpoint_node_label, ("Switch", "Cluster", "UserEquipment"), graph)
        if startpoint_node is None:
            raise HTTPNotFound("Starting point node {0} doesn't exist.".format(startpoint_node_label))

        endpoint_node = all_nodes.match("Endpoint", name=endpoint_node_label, uuid=graph_id).first()
        if endpoint_node is None:
            raise HTTPNotFound("Endpoint node {0} doesn't exist.".format(endpoint_node_label))

        # check if the endpoint is hosted by the compute node before running the RTT cypher query
        hosted_by_node = graph.relationships.match(nodes=(endpoint_node, None), r_type="hostedBy").first().end_node
        if hosted_by_node["name"] == startpoint_node["name"]:
            result = {"forward_latencies": [], "reverse_latencies": [], "response_time": endpoint_node["response_time"],
                      "request_size": endpoint_node["request_size"], "response_size": endpoint_node["response_size"]}
        else:
            query_to_execute = RTT_CYPHER_QUERY_TEMPLATE.format(startpoint_node_type, startpoint_node_label, endpoint_node_label, graph_id)
            log.info("Executing cypher query: {0}".format(query_to_execute))
            data = graph.run(query_to_execute).data()  # returns a list of dictionaries, each dictionary represents a row in the result
            try:
                result = data[0]
            except Exception as e:
                log.error("Unexpected error occurred while executing RTT cypher query for graph with UUID {0} - {1}".format(graph_id, e))
                raise HTTPBadRequest("The Neo4j cypher query didn't return a valid result for the temporal graph with ID {0}".format(graph_id))

        sf_edge = graph.match(nodes=(None, endpoint_node), r_type="realisedBy").first()
        sf_node = sf_edge.start_node if sf_edge is not None else None
        if sf_node is None:
            msg = "No service function found associated with endpoint {0}".format(endpoint_node["name"])
            log.error("Unexpected error: {0}".format(msg))
            raise HTTPBadRequest(msg)

        sf_package_edge = graph.match(nodes=(sf_node, None), r_type="instanceOf").first()
        sf_package_node = sf_package_edge.end_node if sf_package_edge is not None else None
        if sf_package_node is None:
            msg = "No service function package found associated with service function {0}".format(sf_node["name"])
            log.error("Unexpected error: {0}".format(msg))
            raise HTTPBadRequest(msg)

        result["global_tags"] = {"flame_sfe": endpoint_node["name"], "flame_server": hosted_by_node["name"], "flame_location": hosted_by_node["name"],
                                 "flame_sfc": reference_node["sfc"], "flame_sfci": reference_node["sfci"], "flame_sfp": sf_package_node["name"], "flame_sf": sf_node["name"]}
        result["local_tags"] = {"traffic_source": startpoint_node_label}

        # calculate the Round-Trip-Time
        total_forward_latency = sum(result["forward_latencies"])
        result["total_forward_latency"] = total_forward_latency
        total_reverse_latency = sum(result["reverse_latencies"])
        result["total_reverse_latency"] = total_reverse_latency
        service_delay = result["response_time"]

        # the service delay is assumed to include the network data delay, so no need to include the bandwidth and the request/response size in the calculation
        # bandwidth = self.request.registry.settings["network_bandwidth"]
        # request_size = result["request_size"]
        # response_size = result["response_size"]
        # round_trip_time = self.calculate_round_trip_time(total_forward_latency, total_reverse_latency, service_delay, request_size, response_size, bandwidth)

        # the calculation in the new model is simple the network latency added to the service processing time (network data delay + any time the service spends on processing)
        round_trip_time = total_forward_latency + service_delay + total_reverse_latency
        result["round_trip_time"] = round_trip_time

        return result

    # @staticmethod
    # def calculate_round_trip_time(forward_latency, reverse_latency, service_delay, request_size, response_size, bandwidth, packet_size=1500, packet_header_size=50):
    #     """
    #     Calculates the round trip time given the list of arguments.
    #
    #     (DEPRECATED - this is a very simplistic model which uses a fixed static bandwidth value and fixed packet / packet header size)
    #
    #     :param forward_latency: network latency in forward direction (s)
    #     :param reverse_latency: network latency in reverse direction (s)
    #     :param service_delay: media service delay (s)
    #     :param request_size: request size (bytes)
    #     :param response_size: response size (bytes)
    #     :param bandwidth: network bandwidth (Mb/s)
    #     :param packet_size: size of packet (bytes)
    #     :param packet_header_size: size of the header of the packet (bytes)
    #     :return: the calculated round trip time
    #     """
    #
    #     forward_data_delay = (8/10**6) * (request_size / bandwidth) * (packet_size / (packet_size - packet_header_size))
    #     reverse_data_delay = (8/10**6) * (response_size / bandwidth) * (packet_size / (packet_size - packet_header_size))
    #
    #     rtt = forward_latency + forward_data_delay + service_delay + reverse_latency + reverse_data_delay
    #
    #     return rtt

    @view_config(route_name='graph_network_topology', request_method='POST')
    def build_network_topology(self):
        """
        An API endpoint to build the network topology in the neo4j graph, however, without updating the latency of existing edges.

        :return: A JSON response with the number of switches, clusters and ues that were built.
        """

        return self._build_network_topology(update_existing_links=False)

    @view_config(route_name='graph_network_topology', request_method='PUT')
    def build_and_update_network_topology(self):
        """
        An API endpoint to build the network topology in the neo4j graph and also update the latency of existing edges.

        :return: A JSON response with the number of switches, clusters and ues that were built.
        """

        return self._build_network_topology(update_existing_links=True)

    def _build_network_topology(self, update_existing_links):
        """
        A utility method used to build the network graph topology with the option of updating existing links.

        :param update_existing_links: a flag to be set to True if existing network links need to be updated

        :return: A dictionary with the number of switches, clusters and ues that were created.
        """

        graph = self.get_graph_reference()

        sdn_controller_ip = self.request.registry.settings['sdn_controller_ip']
        sdn_controller_port = self.request.registry.settings['sdn_controller_port']

        # retrieve all switches
        url = "http://{0}:{1}{2}".format(sdn_controller_ip, sdn_controller_port, "/wm/core/controller/switches/json")
        log.info("Fetching the list of switches from the SDN controller - {0}".format(url))
        switches_json_response = self.get_sdn_controller_response(url)

        # map the DPID of each switch to its IP address
        switches = {}
        for switch in switches_json_response:
            # map the dpid to the switch IP address, the IP address is in the format '/172.168.23.54:1234'
            switches[switch["switchDPID"]] = switch["inetAddress"][1:].split(":")[0]

        # retrieve all external links (gathered through BDDP)
        url = "http://{0}:{1}{2}".format(sdn_controller_ip, sdn_controller_port, "/wm/topology/external-links/json")
        log.info("Fetching all external links from the SDN controller - {0}".format(url))
        external_links = self.get_sdn_controller_response(url)

        # retrieve all local links (gathered through LLDP)
        url = "http://{0}:{1}{2}".format(sdn_controller_ip, sdn_controller_port, "/wm/topology/links/json")
        log.info("Fetching all local links from the SDN controller - {0}".format(url))
        local_links = self.get_sdn_controller_response(url)

        # TODO this is a temporary solution - currently the service router to cluster mapping is read from a file (which must be manually prepared beforehand)
        clusters_file = self.request.registry.settings["network_clusters_path"]
        try:
            with open(clusters_file) as fh:
                clusters = load(fh)
        except Exception as e:
            log.error("Unexpected error: {0}".format(e))
            log.error("No service-router-to-cluster mapping was found while building the network topology.")
            clusters = {}

        # TODO this is a temporary solution - currently the service router to ue mapping is read from a file (which must be manually prepared beforehand)
        ues_file = self.request.registry.settings["network_ues_path"]
        try:
            with open(ues_file) as fh:
                ues = load(fh)
        except Exception as e:
            log.error("Unexpected error: {0}".format(e))
            log.error("No service-router-to-ue mapping was found while building the network topology.")
            ues = {}

        # build the network graph and retrieve the number of switch nodes and cluster nodes that were created
        tmp_switch_count, tmp_clusters_count, tmp_ues_count = build_network_graph(graph, switches, external_links, clusters, ues, update_existing_links=update_existing_links)
        switch_count, clusters_count, ues_count = build_network_graph(graph, switches, local_links, clusters, ues, update_existing_links=update_existing_links)
        switch_count += tmp_switch_count
        clusters_count += tmp_clusters_count
        ues_count += tmp_ues_count

        return {"new_switches_count": switch_count, "new_clusters_count": clusters_count, "new_ues_count": ues_count}

    @staticmethod
    def get_sdn_controller_response(url):
        """
        Send a GET request to the SDN controller and validate the response - expecting a JSON response (currently only applicable for Floodlight)

        :param url: the full url - SDN controller IP and port + API endpoint path

        :return: the response in JSON

        :raises HTTPServiceUnavailable: if the SDN controller couldn't be reached on the given IP and port
        :raises HTTPNotImplemented: if the SDN controller returned status code different than 200 or the response couldn't be converted to JSON
        """

        # try getting any response from the controller
        try:
            response = get(url)
        except exceptions.ConnectionError:
            msg = "The SDN controller is not available - request was sent to {0}.".format(url)
            log.error("Unexpected error: {0}".format(msg))
            raise HTTPServiceUnavailable("The SDN controller couldn't be reached when trying to build the network topology.")

        # check if the SDN controller returned the expected response
        if response.status_code != 200:
            msg = "The SDN controller returned a response with status code different than 200 - request was sent to {0}.".format(url)
            log.error("Unexpected error: {0}".format(msg))
            raise HTTPNotImplemented("The SDN controller failed to return a successful response status code.")

        # check if the SDN controller returned a JSON response
        try:
            response_json = response.json()
        except ValueError:  # response not in JSON
            msg = "The SDN controller returned a response which couldn't be converted to JSON - request was sent to {0}.".format(url)
            log.error("Unexpected error: {0}".format(msg))
            raise HTTPNotImplemented("The SDN controller failed to return a valid JSON response.")

        return response_json

    @view_config(route_name='graph_network_topology', request_method='DELETE')
    def delete_network_topology(self):
        """
        An API endpoint to delete the network topology in the neo4j graph.

        :return: A JSON response with the number of switches, clusters and ues that were deleted.
        """

        graph = self.get_graph_reference()

        deleted_switches, deleted_clusters, deleted_ues = delete_network_graph(graph)

        return {"deleted_switches_count": deleted_switches, "deleted_clusters_count": deleted_clusters, "deleted_ues_count": deleted_ues}

    @view_config(route_name='graph_execute_pipeline', request_method='POST')
    def execute_graph_pipeline(self):
        """
        An API endpoint to execute the graph pipeline script as a background process.

        :return: A JSON response with the database name and request's uuid.
        """

        try:
            body = self.request.body.decode(self.request.charset)
            json_queries = validate_monitor_request_body(body)  # validate the content and receive a json dictionary object
        except AssertionError as e:
            raise HTTPBadRequest("Bad request content: {0}".format(e.args))

        influx_client = InfluxDBClient(host=self.request.registry.settings['influx_host'], port=self.request.registry.settings['influx_port'], timeout=10)

        sfc = json_queries["service_function_chain"]
        database_name = json_queries["service_function_chain"]  # currently DB is named after the SFC
        if database_name not in [db["name"] for db in influx_client.get_list_database()]:
            raise HTTPBadRequest("Database for service function chain {0} not found.".format(database_name))

        # get the list of UEs from the Neo4j graph - if no UEs are found, the network topology has not been built yet, so return bad request
        graph = self.get_graph_reference()

        ue_nodes = graph.nodes.match("UserEquipment")
        ues_list = [ue_node["name"] for ue_node in ue_nodes]
        # return bad request for empty list of UEs
        if len(ues_list) == 0:
            raise HTTPBadRequest("Graph pipeline process for SFC {0} cannot be started - no UserEquipment nodes found, the graph network topology has not been built.".format(sfc))
        # add the list of UEs to the JSON configuration passed to the pipeline script
        json_queries["ues"] = ues_list

        # start a graph pipeline subprocess
        process = Popen(["graph-pipeline.sh", dumps(json_queries)])
        process_pid = process.pid
        process_return_code = process.returncode

        # check if for some reason the process returned immediately (could be if the graph-pipeline script is not found
        if process_return_code is None:  # process has started running
            request_uuid = str(uuid4())  # generate a request UUID used to map to the process ID of the graph pipeline

            log.info("Started a graph pipeline process for SFC {0} with PID {1} and request UUID {2}".format(sfc, process_pid, request_uuid))

            MonitoringProcess.add({"request_id": request_uuid, "process_id": process_pid})

            return {"database": database_name, "uuid": request_uuid}
        else:  # a valid returned code was returned, hence the process has terminated one way or another - we do not expect this since the pipeline script must be continuously running
            log.warning("Graph pipeline process for SFC {0} with PID {1} has finished executing unexpectedly with return code {2}".format(sfc, process_pid, process_return_code))
            raise HTTPInternalServerError("An unexpected error occurred while trying to start monitoring graph measurements for service function chain {0}".format(sfc))

    @view_config(route_name='graph_manage_pipeline', request_method='GET')
    def get_graph_pipeline_status(self):
        """
        An API endpoint to get the status of a monitoring graph pipeline script.

        :return: A JSON response with the status of the background process.
        """

        request_id = self.request.matchdict['request_id']  # get the UUID of the request from the URL
        process_id = MonitoringProcess.get(request_id)

        if process_id is None:
            raise HTTPNotFound("A monitoring process with ID {0} couldn't be found.".format(request_id))

        # create a process management class instance
        try:
            process_obj = Process(process_id)
            status = process_obj.status()
            log.info("Fetching process status with request ID {0} and process ID {1}, status - {2}".format(request_id, process_id, status))
            response = {"status": status, "msg": "Successfully fetched status of graph pipeline process."}
        except NoSuchProcess as e:
            log.warning("Unexpected error occurred when trying to get a process that is registered in the CLMC service, but doesn't exist on the OS - {0}".format(e))
            response = {"msg": "Monitoring process has been stopped or killed or terminated before this request was executed."}

        return response

    @view_config(route_name='graph_manage_pipeline', request_method='DELETE')
    def stop_graph_pipeline(self):
        """
        An API endpoint to stop a running in the background graph pipeline script.

        :return: A JSON response with a simple info message for the process
        """

        request_id = self.request.matchdict['request_id']  # get the UUID of the request from the URL
        process_id = MonitoringProcess.get(request_id)

        if process_id is None:
            raise HTTPNotFound("A monitoring process with ID {0} couldn't be found.".format(request_id))

        # create a process management class instance and terminate the process
        try:
            process_obj = Process(process_id)
            log.info("Terminating process with request ID {0} and process ID {1}, status before termination - {2}".format(request_id, process_id, process_obj.status()))
            process_obj.terminate()
            process_obj.wait(timeout=3)
            response = {"msg": "Monitoring process has been successfully stopped."}
        except TimeoutExpired as e:
            log.error("Unexpected error occurred while waiting for a graph pipeline process to terminate - {0}".format(e))
            raise HTTPInternalServerError("Termination of monitoring process couldn't be completed.")
        except NoSuchProcess as e:
            log.warning("Unexpected error occurred when trying to get a process that is registered in the CLMC service, but doesn't exist on the OS - {0}".format(e))
            response = {"msg": "Monitoring process has been stopped or killed or terminated before this request was executed."}

        MonitoringProcess.delete(request_id)

        return response

    def get_graph_reference(self):
        """
        A utility function to get a Graph DB object reference and clear the cache of py2neo since it might lead to erroneous behaviour

        :return: reference to the graph DB object
        """

        graph = Graph(host=self.request.registry.settings['neo4j_host'], password=self.request.registry.settings['neo4j_password'])

        # TODO we probably need to remove the caching functionality of py2neo and maintain our own version
        # NOTE: make sure Py2neo caching is disabled!
        graph.node_cache.clear()
        graph.relationship_cache.clear()

        return graph
